@@ -1,7 +1,9 @@
 /*
  * ProFTPD - FTP server daemon
- * Copyright (c) 1997, Public Flood Software
- *  
+ * Copyright (c) 1997, 1998 Public Flood Software
+ * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
+ * Copyright (c) 2001, 2002 The ProFTPD Project team
+ *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
@@ -14,7 +16,12 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ *
+ * As a special exemption, Public Flood Software/MacGyver aka Habeeb J. Dihu
+ * and other respective copyright holders give permission to link this program
+ * with OpenSSL, and distribute the resulting executable, without including
+ * the source code for OpenSSL in the source distribution.
  */
 
 #ifndef __TIMERS_H
@@ -26,7 +33,7 @@ typedef struct timer_struc {
   long count;			/* Amount of time remaining */
   long interval;		/* Original length of timer */
 
-  int timerno;			/* Caller dependant timer number */
+  int timerno;			/* Caller dependent timer number */
   module *mod;			/* Module owning this timer */
   callback_t callback;		/* Function to callback */
   char remove;			/* Internal use */
@@ -34,9 +41,10 @@ typedef struct timer_struc {
 
 
 /* Prototypes */
-int add_timer(int seconds,int timerno,module *mod, callback_t cb);
-int remove_timer(int timerno,module *mod);
-int reset_timer(int timerno,module *mod);
+int add_timer(int, int, module *, callback_t);
+int remove_timer(int, module *);
+int reset_timer(int, module *);
 int timer_sleep(int);
+void handle_alarm(void);
 
 #endif /* __TIMERS_H */
