@@ -1,7 +1,8 @@
 /*
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
- * Copyright (C) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
+ * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
+ * Copyright (c) 2001, 2002 The ProFTPD Project team
  *  
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,7 +26,7 @@
 
 /*
  * Flexible logging module for proftpd
- * $Id: mod_log.c,v 1.20 2001/10/18 17:10:47 flood Exp $
+ * $Id: mod_log.c,v 1.26 2002/08/14 16:25:29 castaglia Exp $
  */
 
 #include "conf.h"
@@ -35,6 +36,7 @@
 extern response_t *resp_list,*resp_err_list;
 
 #define LOGBUF_SIZE		1025
+#define LOG_EXTENDED_MODE	0644
 
 typedef struct logformat_struc	logformat_t;
 typedef struct logfile_struc 	logfile_t;
@@ -123,8 +125,8 @@ static void add_meta(unsigned char **s, unsigned char meta, int args,
     while(args--) {
       arglen = va_arg(ap, int);
       arg = va_arg(ap, char *);
-      
-      bcopy(arg, *s, arglen);
+     
+      memcpy(*s, arg, arglen); 
       (*s) = (*s) + arglen;
       **s = META_ARG_END;
       (*s) = (*s) + 1;
@@ -246,7 +248,7 @@ void logformat(char *nickname, char *fmts)
   lf = (logformat_t *) pcalloc(log_pool, sizeof(logformat_t));
   lf->lf_nickname = pstrdup(log_pool, nickname);
   lf->lf_format = palloc(log_pool, outs - format);
-  bcopy(format, lf->lf_format, outs - format);
+  memcpy(lf->lf_format, format, outs - format);
   
   if(!format_set)
     format_set = xaset_create(log_pool, NULL);
@@ -257,9 +259,7 @@ void logformat(char *nickname, char *fmts)
 
 /* Syntax: LogFormat nickname "format string"
  */
-
-MODRET add_logformat(cmd_rec *cmd)
-{
+MODRET add_logformat(cmd_rec *cmd) {
   CHECK_ARGS(cmd, 2);
   CHECK_CONF(cmd, CONF_ROOT);
 
@@ -307,14 +307,12 @@ static int _parse_classes(char *s)
 
 /* Syntax: ExtendedLog <log-filename> [<cmd-classes> [<format-nickname>]]
  */
-
-MODRET add_extendedlog(cmd_rec *cmd)
-{
-  config_rec *c;
+MODRET add_extendedlog(cmd_rec *cmd) {
+  config_rec *c = NULL;
   int argc;
   char **argv;
 
-  CHECK_CONF(cmd,CONF_ROOT|CONF_VIRTUAL|CONF_ANON|CONF_GLOBAL);
+  CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL|CONF_ANON);
 
   argc = cmd->argc;
   argv = cmd->argv;
@@ -323,7 +321,8 @@ MODRET add_extendedlog(cmd_rec *cmd)
     CONF_ERROR(cmd, "Syntax: ExtendedLog <log-filename> "
 	       "[<Command-Classes> [<Format-Nickname>]]");
 
-  c = add_config_param("ExtendedLog",3,NULL,NULL,NULL);
+  c = add_config_param(cmd->argv[0], 3, NULL, NULL, NULL);
+  c->flags |= CF_MERGEDOWN;
 
   if(cmd->argv[1][0] != '/')
     c->argv[0] = dir_canonical_path(log_pool,cmd->argv[1]);
@@ -340,9 +339,8 @@ MODRET add_extendedlog(cmd_rec *cmd)
 }
 
 /* Syntax: AllowLogSymlinks <on|off> */
-
 MODRET set_allowlogsymlinks(cmd_rec *cmd) {
-  int bool;
+  int bool = -1;
 
   CHECK_ARGS(cmd, 1);
   CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL);
@@ -350,55 +348,53 @@ MODRET set_allowlogsymlinks(cmd_rec *cmd) {
   if ((bool = get_boolean(cmd, 1)) == -1)
     CONF_ERROR(cmd, "expected boolean argument.");
 
-  add_config_param("AllowLogSymlinks", 1, (void *) bool);
+  add_config_param(cmd->argv[0], 1, (void *) bool);
 
   return HANDLED(cmd);
 }
 
 /* Syntax: SystemLog <filename> */
-
-MODRET set_systemlog(cmd_rec *cmd)
-{
-  char *syslogfn;
+MODRET set_systemlog(cmd_rec *cmd) {
+  char *syslogfn = NULL;
   int ret;
   
-  CHECK_ARGS(cmd,1);
-  CHECK_CONF(cmd,CONF_ROOT);
+  CHECK_ARGS(cmd, 1);
+  CHECK_CONF(cmd, CONF_ROOT);
 
   log_closesyslog();
 
   syslogfn = cmd->argv[1];
 
-  if(strcasecmp(syslogfn,"NONE") == 0) {
+  if (strcasecmp(syslogfn, "NONE") == 0) {
     log_discard();
     return HANDLED(cmd);
   }
 
-  if(*syslogfn != '/') 
+  if (*syslogfn != '/') 
     syslogfn = dir_canonical_path(cmd->tmp_pool,syslogfn);
 
   block_signals();
   PRIVS_ROOT
 
-    if((ret = log_opensyslog(syslogfn)) < 0) {
-      int xerrno = errno;
+  if ((ret = log_opensyslog(syslogfn)) < 0) {
+    int xerrno = errno;
       
-      PRIVS_RELINQUISH
-      unblock_signals();
-      
-      /* NB: these negative values really should be #defined */
-      
-      if(ret == -2) {
-	CONF_ERROR(cmd, "you are attempting to log to a world writeable directory");
+    PRIVS_RELINQUISH
+    unblock_signals();
+    
+    if (ret == LOG_WRITEABLE_DIR) {
+      CONF_ERROR(cmd,
+        "you are attempting to log to a world writeable directory");
 
-      } else if (ret == -3) {
-        CONF_ERROR(cmd, "you are attempting to log to a symbolic link");
+    } else if (ret == LOG_SYMLINK) {
+      CONF_ERROR(cmd, "you are attempting to log to a symbolic link");
 
-      } else {
-	CONF_ERROR(cmd,pstrcat(cmd->tmp_pool,"unable to redirect logging to '",
-			       syslogfn,"': ",strerror(xerrno),NULL));
-      }
+    } else {
+      CONF_ERROR(cmd, pstrcat(cmd->tmp_pool,
+        "unable to redirect logging to '", syslogfn, "': ",
+        strerror(xerrno), NULL));
     }
+  }
   
   PRIVS_RELINQUISH
   unblock_signals();
@@ -474,8 +470,8 @@ char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f)
     
   case META_BYTES_SENT:
     argp = arg;
-    if(session.xfer.p)
-      snprintf(argp, sizeof(arg), "%lu",session.xfer.total_bytes);
+    if (session.xfer.p)
+      snprintf(argp, sizeof(arg), "%" PR_LU, session.xfer.total_bytes);
     else
       sstrncpy(argp, "-", sizeof(arg));
 
@@ -606,8 +602,8 @@ char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f)
         end_time.tv_sec--;
       }
 
-      snprintf(argp, sizeof(arg), "%lu.%03lu", (unsigned long) end_time.tv_sec,
-	       (unsigned long) (end_time.tv_usec / 1000));
+      snprintf(argp, sizeof(arg), "%ld.%03ld", (time_t) end_time.tv_sec,
+	       (time_t) (end_time.tv_usec / 1000));
     } else {
       sstrncpy(argp,"-",sizeof(arg));
     }
@@ -705,7 +701,7 @@ void do_log(cmd_rec *cmd, logfile_t *lf)
         if(tmp > size)
           tmp = size;
 
-        bcopy(s,bp,tmp);
+        memcpy(bp, s, tmp);
         size -= tmp;
         bp += tmp;
       }
@@ -755,7 +751,7 @@ void log_rehash(void *d)
 }
   
 static 
-int log_init()
+int log_init(void)
 {
   log_pool = make_sub_pool(permanent_pool);
   /* add the "default" extendedlog format */
@@ -764,7 +760,7 @@ int log_init()
   return 0;
 }
 
-static void get_extendedlogs() {
+static void get_extendedlogs(void) {
   config_rec *c;
   char *logfname;
   int logclasses = CL_ALL;
@@ -873,93 +869,41 @@ MODRET log_auth_complete(cmd_rec *cmd)
 }
 
 static
-int log_child_init()
+int log_child_init(void)
 {
   /* open all log files */
-  logfile_t *lf;
-  int log_open = FALSE;
+  logfile_t *lf = NULL;
 
   get_extendedlogs();
 
-  for(lf = logs; lf; lf=lf->next) {
-    char *ptr;
-    struct stat statbuf;
+  for (lf = logs; lf; lf = lf->next) {
+
+    if (lf->lf_fd == -1) {
+      int res = 0;
 
-    if(lf->lf_fd == -1) {
-      log_open = FALSE;
       block_signals();
       PRIVS_ROOT
-
-      /* make sure we're not logging to a world-writeable directory, and
-       * that we're not logging to a symlink
-       */
-      if ((ptr = rindex(lf->lf_filename, '/')) != NULL) {
-        *ptr = '\0';
-
-        if (stat(lf->lf_filename, &statbuf) == -1) {
-          PRIVS_RELINQUISH
-          unblock_signals();
-          log_debug(DEBUG0, "error: stat(%s): %s", lf->lf_filename,
-            strerror(errno));
-          continue;
-        }
-
-        if (!S_ISDIR(statbuf.st_mode)) {
-          PRIVS_RELINQUISH
-          unblock_signals();
-          log_debug(DEBUG0, "%s is not a directory", lf->lf_filename);
-          continue;
-        }
-
-        if (statbuf.st_mode & S_IWOTH) {
-          PRIVS_RELINQUISH
-          unblock_signals();
-          log_debug(DEBUG0, "%s is a world writeable directory",
-            lf->lf_filename);
-          continue;
-        }
-
-        *ptr = '/';
-      }
-
-      if (get_param_int(main_server->conf, "AllowLogSymlinks", FALSE) != TRUE) {
-
-        /* prevent a race condition between stat() and open() by opening the
-         * file now, _then_ checking to see if it's a symlink
-         */
-        if ((lf->lf_fd = open(lf->lf_filename, O_APPEND|O_CREAT|O_WRONLY,
-            0644)) == -1) {
-      PRIVS_RELINQUISH
-          unblock_signals();
-          log_pri(LOG_NOTICE, "Unable to open ExtendedLog '%s': %s",
-            lf->lf_filename, strerror(errno));
-          continue;
-        }
-
-        if (lstat(lf->lf_filename, &statbuf) != -1 &&
-            S_ISLNK(statbuf.st_mode)) {
-          PRIVS_RELINQUISH
-          unblock_signals();
-          log_pri(LOG_NOTICE,
-            "Unable to open ExtendedLog '%s': %s is a symbolic link",
-            lf->lf_filename, lf->lf_filename);
-          close(lf->lf_fd);
-          lf->lf_fd = -1;
-          continue;
-        }
-
-        log_open = TRUE;
-      }
-
-      if (!log_open)
-        lf->lf_fd = open(lf->lf_filename,O_CREAT|O_APPEND|O_WRONLY,0644);
-
+      res = log_openfile(lf->lf_filename, &lf->lf_fd, LOG_EXTENDED_MODE);
       PRIVS_RELINQUISH
       unblock_signals();
 
-      if(lf->lf_fd == -1)
-        log_pri(LOG_NOTICE, "Unable to open ExtendedLog '%s': %s",
-                lf->lf_filename, strerror(errno));
+      if (res == -1) {
+        log_pri(LOG_NOTICE, "unable to open ExtendedLog '%s': %s",
+          lf->lf_filename, strerror(errno));
+        continue;
+
+      } else if (res == LOG_WRITEABLE_DIR) {
+        log_pri(LOG_NOTICE, "unable to open ExtendedLog '%s': "
+          "containing directory is world writeable", lf->lf_filename);
+        continue;
+
+      } else if (res == LOG_SYMLINK) {
+        log_pri(LOG_NOTICE, "unable to open ExtendedLog '%s': "
+          "%s is a symbolic link", lf->lf_filename, lf->lf_filename);
+        close(lf->lf_fd);
+        lf->lf_fd = -1;
+        continue;
+      }
     }
   }
 
@@ -967,7 +911,7 @@ int log_child_init()
 }
 
 
-static conftable log_config[] = {
+static conftable log_conftab[] = {
   { "AllowLogSymlinks",	set_allowlogsymlinks,			NULL },
   { "LogFormat",	add_logformat,				NULL },
   { "ExtendedLog",	add_extendedlog,			NULL },
@@ -975,7 +919,7 @@ static conftable log_config[] = {
   { NULL,		NULL,					NULL }
 };
 
-static cmdtable log_commands[] = {
+static cmdtable log_cmdtab[] = {
   { PRE_CMD, 		C_QUIT,	G_NONE,	log_command,		FALSE, FALSE },
   { LOG_CMD,		C_ANY,	G_NONE,	log_command,		FALSE, FALSE },
   { LOG_CMD_ERR,	C_ANY,	G_NONE,	log_command,		FALSE, FALSE },
@@ -987,8 +931,9 @@ module log_module = {
   NULL,NULL,			/* Always NULL */
   0x20,				/* API version */
   "log",			/* Module name */
-  log_config,
-  log_commands,
+  log_conftab,
+  log_cmdtab,
   NULL,
-  log_init,log_child_init
+  log_init,
+  log_child_init
 };
