@@ -1,7 +1,8 @@
 /*
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
- * Copyright (C) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
+ * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
+ * Copyright (c) 2001, 2002 The ProFTPD Project team
  *  
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -24,7 +25,7 @@
  */
 
 /* Directory listing module for ProFTPD.
- * $Id: mod_ls.c,v 1.45 2001/09/26 15:32:47 flood Exp $
+ * $Id: mod_ls.c,v 1.61 2002/08/14 16:25:29 castaglia Exp $
  */
 
 #include "conf.h"
@@ -44,14 +45,17 @@ static int listdir(cmd_rec*, pool*, const char *name);
 
 static int matches = 0;
 static char *default_options;
-static int showsymlinks, showsymlinks_hold, timesgmt = 0;
+static unsigned char list_show_symlinks = TRUE, list_times_gmt = FALSE;
+static unsigned char show_symlinks_hold;
 static int cmp(const void *a, const void *b);
 static char *fakeuser, *fakegroup;
-static umode_t fakemode;
+static mode_t fakemode;
 static int fakemodep;
 static int ls_errno = 0;
 static time_t ls_curtime = 0;
 
+static unsigned char use_globbing = TRUE;
+
 /* ls options */
 static int
     opt_a = 0,
@@ -61,6 +65,7 @@ static int
     opt_F = 0,
     opt_l = 0,
     opt_L = 0,
+    opt_n = 0,
     opt_R = 0,
     opt_r = 0,
     opt_t = 0,
@@ -118,27 +123,26 @@ static config_rec *_find_ls_limit(char *ftp_cmd) {
   return NULL;
 }
 
-static void push_cwd(char *_cwd, int *symhold) {
-  if(!_cwd)
+static void push_cwd(char *_cwd, unsigned char *symhold) {
+  if (!_cwd)
     _cwd = cwd;
 
-  if(!symhold)
-    symhold = &showsymlinks_hold;
+  if (!symhold)
+    *symhold = show_symlinks_hold;
 
   sstrncpy(_cwd, fs_getcwd(), MAXPATHLEN + 1);
-  *symhold = showsymlinks;
+  *symhold = list_show_symlinks;
 }
 
-static void pop_cwd(char *_cwd, int *symhold)
-{
-  if(!_cwd)
+static void pop_cwd(char *_cwd, unsigned char *symhold) {
+  if (!_cwd)
     _cwd = cwd;
 
-  if(!symhold)
-    symhold = &showsymlinks_hold;
+  if (!symhold)
+    *symhold = show_symlinks_hold;
 
-  fs_chdir(_cwd,*symhold);
-  showsymlinks = *symhold;
+  fs_chdir(_cwd, *symhold);
+  list_show_symlinks = *symhold;
 }
 
 static int ls_perms_full(pool *p, cmd_rec *cmd, const char *path, int *hidden)
@@ -162,17 +166,17 @@ static int ls_perms_full(pool *p, cmd_rec *cmd, const char *path, int *hidden)
   else
     ret = dir_check(p,cmd->argv[0],cmd->group,fullpath,hidden);
   
-  if(session.dir_config) {
-    showsymlinks = get_param_int(session.dir_config->subset,
-                                 "ShowSymlinks",FALSE);
+  if (session.dir_config) {
+    unsigned char *tmp = get_param_ptr(session.dir_config->subset,
+      "ShowSymlinks", FALSE);
 
-    if(showsymlinks == -1)
-      showsymlinks = 1;
+    if (tmp)
+      list_show_symlinks = *tmp;
   }
 
   _fakemode = get_param_int(CURRENT_CONF,"DirFakeMode",FALSE);
   if(_fakemode != -1) {
-    fakemode = (umode_t)_fakemode;
+    fakemode = (mode_t)_fakemode;
     fakemodep = 1;
   } else
     fakemodep = 0;
@@ -201,17 +205,17 @@ static int ls_perms(pool *p, cmd_rec *cmd, const char *path,int *hidden)
   
   ret = dir_check(p,cmd->argv[0],cmd->group,fullpath,hidden);
 
-  if(session.dir_config) {
-    showsymlinks = get_param_int(session.dir_config->subset,
-                                 "ShowSymlinks",FALSE);
+  if (session.dir_config) {
+    unsigned char *tmp = get_param_ptr(session.dir_config->subset,
+      "ShowSymlinks",FALSE);
 
-    if(showsymlinks == -1)
-      showsymlinks = 1;
+    if (tmp)
+      list_show_symlinks = *tmp;
   }
 
   _fakemode = get_param_int(CURRENT_CONF,"DirFakeMode",FALSE);
   if(_fakemode != -1) {
-    fakemode = (umode_t)_fakemode;
+    fakemode = (mode_t)_fakemode;
     fakemodep = 1;
   } else
     fakemodep = 0;
@@ -266,7 +270,7 @@ int listfile(cmd_rec *cmd, pool *p, const char *name)
   if(fs_lstat(name,&st) == 0) {
     suffix[0] = suffix[1] = '\0';
 
-    if(S_ISLNK(st.st_mode) && (opt_L || !showsymlinks)) {
+    if(S_ISLNK(st.st_mode) && (opt_L || !list_show_symlinks)) {
       /* attempt to fully dereference symlink */
       struct stat l_st;
 
@@ -303,12 +307,13 @@ int listfile(cmd_rec *cmd, pool *p, const char *name)
       return 0;
 
     mtime = st.st_mtime;
-    if(timesgmt)
-      t = gmtime((time_t*) &mtime);
+
+    if (list_times_gmt)
+      t = gmtime((time_t *) &mtime);
     else 
-      t = localtime((time_t*) &mtime);
+      t = localtime((time_t *) &mtime);
     
-    if(!t) {
+    if (!t) {
       add_response_err(R_421,"Fatal error (localtime() returned NULL?!?)");
       return -1;
     }
@@ -353,7 +358,7 @@ int listfile(cmd_rec *cmd, pool *p, const char *name)
       if(m[0] != ' ') {
         char nameline[MAXPATHLEN + MAXPATHLEN + 128] = {'\0'};
         char timeline[6] = {'\0'};
-        umode_t mode = st.st_mode;
+        mode_t mode = st.st_mode;
 
         if(fakemodep) {
           mode = fakemode;
@@ -396,11 +401,23 @@ int listfile(cmd_rec *cmd, pool *p, const char *name)
         else
           snprintf(timeline, sizeof(timeline), "%02d:%02d",t->tm_hour,t->tm_min);
 
-        snprintf(nameline, sizeof(nameline), "%s %3d %-8s %-8s %8d %s %2d %s %s", m,
-                (int)st.st_nlink, MAP_UID(st.st_uid), 
-                MAP_GID(st.st_gid),
-                (unsigned int)st.st_size, months[t->tm_mon],
-                t->tm_mday, timeline, name);
+        if (!opt_n) {
+
+          /* Format nameline using user/group names. */
+          snprintf(nameline, sizeof(nameline), "%s %3d %-8s %-8s %8" PR_LU
+            " %s %2d %s %s", m, st.st_nlink, MAP_UID(st.st_uid),
+            MAP_GID(st.st_gid), st.st_size, months[t->tm_mon], t->tm_mday,
+            timeline, name);
+
+        } else {
+
+          /* Format nameline using user/group IDs. */
+          snprintf(nameline, sizeof(nameline), "%s %3d %-8u %-8u %8" PR_LU
+            " %s %2d %s %s", m, st.st_nlink, st.st_uid, st.st_gid,
+            st.st_size, months[t->tm_mon], t->tm_mday,
+            timeline, name);
+
+        }
 
         if(S_ISLNK(st.st_mode)) {
           char *p = nameline + strlen(nameline);
@@ -647,7 +664,7 @@ int outputfiles(cmd_rec *cmd)
 
 
 static
-void discard_output()
+void discard_output(void)
 {
   if(fpool)
     destroy_pool(fpool);
@@ -805,10 +822,10 @@ static int listdir(cmd_rec *cmd, pool *workp, const char *name)
 
   dir = sreaddir(workp,".",TRUE);
 
-	/* search for relevant <Limit>'s to this LIST command.  If found,
-	 * check to see whether hidden files should be ignored
+  /* search for relevant <Limit>'s to this LIST command.  If found,
+   * check to see whether hidden files should be ignored
    */
-	if ((c = _find_ls_limit(cmd->argv[0])) != NULL &&
+  if ((c = _find_ls_limit(cmd->argv[0])) != NULL &&
       get_param_int(c->subset, "IgnoreHidden", FALSE) == TRUE)
     ignore_hidden = TRUE;
 
@@ -830,16 +847,15 @@ static int listdir(cmd_rec *cmd, pool *workp, const char *name)
     s = dir;
     while(*s) {
       if (**s == '.') {
-
-				if (!opt_a && (!opt_A || is_dotdir(*s))) {
-	d = 0;
+        if (!opt_a && (!opt_A || is_dotdir(*s))) {
+          d = 0;
 
         } else {
 
-					/* make sure IgnoreHidden is properly honored.  "." and
+          /* make sure IgnoreHidden is properly honored.  "." and
            * ".." are not to be treated as hidden files, though
-					 */
-					if (is_dotdir(*s) || !ignore_hidden)
+           */
+          if (is_dotdir(*s) || !ignore_hidden)
             d = listfile(cmd,workp,*s);
         }
 
@@ -847,55 +863,60 @@ static int listdir(cmd_rec *cmd, pool *workp, const char *name)
         d = listfile(cmd,workp,*s);
       }
 
-      if(!d)
+      if (!d)
 	*s = NULL;
 
       s++;
     }
 
-    if(outputfiles(cmd) < 0) {
-      if(dest_workp)
+    if (outputfiles(cmd) < 0) {
+      if (dest_workp)
         destroy_pool(workp);
       return -1;
     }
 
     r = dir;
-    while(opt_R && r != s) {
+    while (opt_R && r != s) {
       char cwd[MAXPATHLEN + 1] = {'\0'};
-      int symhold;
+      unsigned char symhold;
 
-      if(*r && (strcmp(*r,".") == 0 || strcmp(*r,"..") == 0)) {
+      if (*r && (strcmp(*r, ".") == 0 || strcmp(*r, "..") == 0)) {
         r++;
         continue;
       }
-      
-      push_cwd(cwd,&symhold);
+
+      /* Add some signal processing to this while loop, as it can
+       * potentially recurse deeply.
+       */
+      pr_handle_signals();      
+
+      push_cwd(cwd, &symhold);
      
-      if(*r && ls_perms_full(workp,cmd,(char*)*r,NULL) &&
-	       !fs_chdir_canon(*r, !opt_L && showsymlinks)) {
+      if (*r && ls_perms_full(workp,cmd,(char*)*r,NULL) &&
+          !fs_chdir_canon(*r, !opt_L && list_show_symlinks)) {
         char *subdir;
 
-	if(strcmp(name,".") == 0)
+	if (strcmp(name,".") == 0)
           subdir = *r;
         else
           subdir = pdircat(workp,name,*r,NULL);
 
-	if(opt_STAT) {
-	  add_response(R_211,"");
-          add_response(R_211,"%s:",subdir);
+	if (opt_STAT) {
+	  add_response(R_211, "%s", "");
+          add_response(R_211, "%s:", subdir);
 
-	} else if(sendline("\n%s:\n",subdir) < 0) {
-          pop_cwd(cwd,&symhold);
+	} else if( sendline("\n%s:\n", subdir) < 0) {
+          pop_cwd(cwd, &symhold);
 
-          if(dest_workp)
+          if (dest_workp)
             destroy_pool(workp);
           return -1;
         }
 
-        if(listdir(cmd,workp,subdir) < 0) {
-          pop_cwd(cwd,&symhold);
+        if (listdir(cmd, workp, subdir) < 0) {
+          pop_cwd(cwd, &symhold);
 
-          if(dest_workp)
+          if (dest_workp)
             destroy_pool(workp);
           return -1;
         }
@@ -906,15 +927,13 @@ static int listdir(cmd_rec *cmd, pool *workp, const char *name)
     }
   }
 
-  if(dest_workp)
+  if (dest_workp)
     destroy_pool(workp);
 
   return 0;
 }
 
-static
-void ls_terminate()
-{
+static void ls_terminate(void) {
   if(!opt_STAT) {
     discard_output();
     if(!XFER_ABORTED) {  /* an error has occured, other than client ABOR */
@@ -953,6 +972,9 @@ void _parse_options(char **opt, int *glob_flags)
       case 'L':
         opt_L++;
         break;
+      case 'n':
+        opt_n = 1;
+        break;
       case '1':
         opt_l = opt_C = 0;
         break;
@@ -985,11 +1007,10 @@ void _parse_options(char **opt, int *glob_flags)
   }
 }
 
-/* The main work for LIST and STAT (not NLST).
-   Return -1 on error, 0 if successful.  */
-static
-int dolist(cmd_rec *cmd, const char *opt, int clearflags)
-{
+/* The main work for LIST and STAT (not NLST).  Returns -1 on error, 0 if
+ * successful.
+ */
+static int dolist(cmd_rec *cmd, const char *opt, int clearflags) {
   int skiparg = 0;
   int glob_flags = GLOB_PERIOD;
   char *arg = (char*)opt;
@@ -998,7 +1019,7 @@ int dolist(cmd_rec *cmd, const char *opt, int clearflags)
   ls_curtime = time(NULL);
 
   if(clearflags) {
-    opt_a = opt_C = opt_d = opt_F = opt_r = opt_R = opt_t = opt_STAT = 0;
+    opt_a = opt_C = opt_d = opt_F = opt_n = opt_r = opt_R = opt_t = opt_STAT = 0;
     opt_L = 0;
   }
 
@@ -1018,7 +1039,10 @@ int dolist(cmd_rec *cmd, const char *opt, int clearflags)
     glob_t g;
     int    a;
     char   pbuffer[MAXPATHLEN + 1] = "";
-      
+     
+    /* make sure the glob_t is initialized */
+    memset(&g, '\0', sizeof(glob_t));
+ 
     if(*arg == '~') {
       struct passwd *pw;
       int i;
@@ -1040,10 +1064,23 @@ int dolist(cmd_rec *cmd, const char *opt, int clearflags)
     if(!ls_perms_full(cmd->tmp_pool, cmd,
 		      (*pbuffer ? (char *) pbuffer : (char *) arg),NULL)) {
       a = -1;
-      skiparg = 1;
+      skiparg = TRUE;
+
     } else {
-      skiparg = 0;
-      a = fs_glob(*pbuffer ? pbuffer : arg, glob_flags, NULL, &g);
+
+      skiparg = FALSE;
+
+      if (use_globbing)
+        a = fs_glob(*pbuffer ? pbuffer : arg, glob_flags, NULL, &g);
+      else {
+
+        /* trick the following code into using the non-glob() processed path */
+        a = 0;
+        g.gl_pathv = (char **) pcalloc(cmd->tmp_pool, 2 * sizeof(char *));
+        g.gl_pathv[0] = (char *) pstrdup(cmd->tmp_pool,
+          *pbuffer ? pbuffer : arg);
+        g.gl_pathv[1] = NULL;
+      }
     }
       
     if(!a) {
@@ -1081,7 +1118,7 @@ int dolist(cmd_rec *cmd, const char *opt, int clearflags)
 	  target_mode = st.st_mode;
 
           if(S_ISLNK(st.st_mode) && (lmode = file_mode((char*)*path)) != 0) {
-            if(opt_L || !showsymlinks)
+            if(opt_L || !list_show_symlinks)
               st.st_mode = lmode;
             target_mode = lmode;
           }
@@ -1089,7 +1126,8 @@ int dolist(cmd_rec *cmd, const char *opt, int clearflags)
           if(opt_d || !(S_ISDIR(target_mode))) {
             if(listfile(cmd,NULL,*path) < 0) {
               ls_terminate();
-	      fs_globfree(&g);
+	      if (use_globbing)
+                fs_globfree(&g);
               return -1;
             }
             **path = '\0';
@@ -1102,7 +1140,8 @@ int dolist(cmd_rec *cmd, const char *opt, int clearflags)
 
       if(outputfiles(cmd) < 0) {
         ls_terminate();
-        fs_globfree(&g);
+        if (use_globbing) 
+          fs_globfree(&g);
 	return -1;
       }
 
@@ -1110,25 +1149,27 @@ int dolist(cmd_rec *cmd, const char *opt, int clearflags)
       while(path && *path) {
         if(**path && ls_perms_full(cmd->tmp_pool,cmd,*path,NULL)) {
           char cwd[MAXPATHLEN + 1] = {'\0'};
-          int symhold;
+          unsigned char symhold;
 
-          if(!justone) {
-            if(opt_STAT) {
-	      add_response(R_211, "");
+          if (!justone) {
+            if (opt_STAT) {
+	      add_response(R_211, "%s", "");
 	      add_response(R_211, "%s:", *path);
+
 	    } else
               sendline("\n%s:\n", *path);
           }
 
-          push_cwd(cwd,&symhold);
+          push_cwd(cwd, &symhold);
 
-          if(!fs_chdir_canon(*path, !opt_L && showsymlinks)) {
+          if(!fs_chdir_canon(*path, !opt_L && list_show_symlinks)) {
             int ret = listdir(cmd, NULL, *path);
             pop_cwd(cwd, &symhold);
 
             if(ret < 0) {
               ls_terminate();
-	      fs_globfree(&g);
+	      if (use_globbing)
+                fs_globfree(&g);
               return -1;
             }
           }
@@ -1136,7 +1177,8 @@ int dolist(cmd_rec *cmd, const char *opt, int clearflags)
 
         if(XFER_ABORTED) {
 	  discard_output();
-          fs_globfree(&g);
+          if (use_globbing)
+            fs_globfree(&g);
 	  return -1;
 	}
 
@@ -1153,7 +1195,7 @@ int dolist(cmd_rec *cmd, const char *opt, int clearflags)
       }
     }
 
-    if(!skiparg)
+    if (!skiparg && use_globbing)
       fs_globfree(&g);
 
     if(XFER_ABORTED) {
@@ -1219,7 +1261,8 @@ static int nlstdir(cmd_rec *cmd, const char *dir) {
        file[MAXPATHLEN + 1] = {'\0'};
   char cwd[MAXPATHLEN + 1]  = {'\0'};
   pool *workp;
-  int curdir = 0, i, symhold, count = 0, hidden = 0;
+  unsigned char symhold;
+  int curdir = 0, i, count = 0, hidden = 0;
   mode_t mode;
   
   config_rec *c = NULL;
@@ -1227,21 +1270,21 @@ static int nlstdir(cmd_rec *cmd, const char *dir) {
 
   workp = make_sub_pool(cmd->tmp_pool);
   
-  if(!*dir || (*dir == '.' && !dir[1]) || strcmp(dir, "./") == 0) {
+  if (!*dir || (*dir == '.' && !dir[1]) || strcmp(dir, "./") == 0) {
     curdir = 1;
     dir = "";
-  } else {
-    push_cwd(cwd,&symhold);
-  }
+
+  } else
+    push_cwd(cwd, &symhold);
   
-  if(fs_chdir_canon(dir, !opt_L && showsymlinks)) {
+  if (fs_chdir_canon(dir, !opt_L && list_show_symlinks)) {
     destroy_pool(workp);
     return 0;
   }
   
-  if((list = sreaddir(workp,".",FALSE)) == NULL) {
-    if(!curdir)
-      pop_cwd(cwd,&symhold);
+  if ((list = sreaddir(workp, ".", FALSE)) == NULL) {
+    if (!curdir)
+      pop_cwd(cwd, &symhold);
     destroy_pool(workp);
     return 0;
   }
@@ -1252,11 +1295,10 @@ static int nlstdir(cmd_rec *cmd, const char *dir) {
   if ((c = _find_ls_limit(cmd->argv[0])) != NULL)
     ignore_hidden = (get_param_int(c->subset, "IgnoreHidden", FALSE) == TRUE);
 
-  while(*list && count >= 0) {
+  while (*list && count >= 0) {
     p = *list; list++;
     
     if (*p == '.') {
-
       if (!opt_a && (!opt_A || is_dotdir(p)))
         continue;
 
@@ -1265,38 +1307,41 @@ static int nlstdir(cmd_rec *cmd, const char *dir) {
       else if (ignore_hidden)
       continue;
     }
-    
-    if((i = fs_readlink(p, file, sizeof(file))) > 0) {
+   
+    if ((i = fs_readlink(p, file, sizeof(file))) > 0) {
       file[i] = '\0';
       f = file;
+
     } else {
       f = p;
     }
     
-    if(ls_perms(workp,cmd,f,&hidden)) {
-      if(hidden)
+    if (ls_perms(workp, cmd, f, &hidden)) {
+      if (hidden)
         continue;
 
       /* If the data connection isn't open, open it now. */
       if((session.flags & SF_XFER) == 0) {
-	if(data_open(NULL,"file list",IO_WRITE,0) < 0) {
+	if(data_open(NULL, "file list", IO_WRITE, 0) < 0) {
 	  data_reset();
 	  count = -1;
 	  continue;
 	}
+
 	session.flags |= SF_ASCII_OVERRIDE;
       }
       
-      if((mode = file_mode(f)) == 0)
+      if ((mode = file_mode(f)) == 0)
 	continue;
       
-      if(!curdir) {
-	if(sendline("%s/%s\n",dir,p) < 0)
+      if (!curdir) {
+	if (sendline("%s/%s\n", dir, p) < 0)
 	  count = -1;
         else
           count++;
+
       } else {
-	if(sendline("%s\n",p) < 0)
+	if (sendline("%s\n", p) < 0)
 	  count = -1;
         else
           count++;
@@ -1304,8 +1349,8 @@ static int nlstdir(cmd_rec *cmd, const char *dir) {
     }
   }
   
-  if(!curdir)
-    pop_cwd(cwd,&symhold);
+  if (!curdir)
+    pop_cwd(cwd, &symhold);
   
   destroy_pool(workp);
   
@@ -1313,15 +1358,13 @@ static int nlstdir(cmd_rec *cmd, const char *dir) {
 }
 
 /* The LIST command.  */
-MODRET genericlist(cmd_rec *cmd)
-{
-  int err;
+MODRET genericlist(cmd_rec *cmd) {
+  int res = 0;
   long _fakemode;
+  unsigned char *tmp = NULL;
 
-  showsymlinks = get_param_int(TOPLEVEL_CONF,"ShowSymlinks",FALSE);
-
-  if(showsymlinks == -1)
-    showsymlinks = 1;
+  if ((tmp = get_param_ptr(TOPLEVEL_CONF, "ShowSymlinks", FALSE)) != NULL)
+    list_show_symlinks = *tmp;
 
   default_options = get_param_ptr(TOPLEVEL_CONF,"LsDefaultOptions",FALSE);
 
@@ -1340,26 +1383,29 @@ MODRET genericlist(cmd_rec *cmd)
     fakegroup = session.group;
 
   _fakemode = (long)get_param_int(TOPLEVEL_CONF,"DirFakeMode",FALSE);
-  timesgmt = get_param_int(TOPLEVEL_CONF, "TimesGMT", FALSE);
-  
-  if((opt_A = get_param_int(TOPLEVEL_CONF,"ShowDotFiles",FALSE)) < 0)
-    opt_A = 0;
-  
-  if(_fakemode != -1) {
-    fakemode = (umode_t)_fakemode;
+
+  if (_fakemode != -1) {
+    fakemode = (mode_t)_fakemode;
     fakemodep = 1;
+
   } else
     fakemodep = 0;
 
-  err = dolist(cmd,cmd->arg,TRUE);
-  if(XFER_ABORTED) {
-    data_abort(0,0);
-    err = -1;
+  if ((tmp = get_param_ptr(TOPLEVEL_CONF, "TimesGMT", FALSE)) != NULL)
+    list_times_gmt = *tmp;
+
+  res = dolist(cmd, cmd->arg, TRUE);
+
+  if (XFER_ABORTED) {
+    data_abort(0, 0);
+    res = -1;
+
   } else if(session.flags & SF_XFER)
     ls_done(cmd);
 
   opt_l = 0;
-  return (err == -1 ? ERROR(cmd) : HANDLED(cmd));
+
+  return (res == -1 ? ERROR(cmd) : HANDLED(cmd));
 }
 
 MODRET fini_nlst(cmd_rec *cmd)
@@ -1368,10 +1414,16 @@ MODRET fini_nlst(cmd_rec *cmd)
   return DECLINED(cmd);
 }
 
-MODRET cmd_stat(cmd_rec *cmd)
+MODRET list_cleanup(cmd_rec *cmd)
 {
+  data_cleanup();
+  return DECLINED(cmd);
+}
+
+MODRET cmd_stat(cmd_rec *cmd) {
   char *arg = cmd->arg;
   long _fakemode;
+  unsigned char *tmp = NULL;
 
   if(cmd->argc < 2) {
     add_response_err(R_500,"'%s' not understood.",get_full_cmd(cmd));
@@ -1385,10 +1437,8 @@ MODRET cmd_stat(cmd_rec *cmd)
   
   while(arg && *arg && isspace((UCHAR)*arg)) arg++;
   
-  showsymlinks = get_param_int(TOPLEVEL_CONF,"ShowSymlinks",FALSE);
-
-  if(showsymlinks == -1)
-    showsymlinks = 1;
+  if ((tmp = get_param_ptr(TOPLEVEL_CONF, "ShowSymlinks", FALSE)) != NULL)
+    list_show_symlinks = *tmp;
 
   default_options = get_param_ptr(TOPLEVEL_CONF,"LsDefaultOptions",FALSE);
 
@@ -1407,16 +1457,16 @@ MODRET cmd_stat(cmd_rec *cmd)
     fakegroup = session.group;
 
   _fakemode = (long)get_param_int(TOPLEVEL_CONF,"DirFakeMode",FALSE);
-  timesgmt = get_param_int(TOPLEVEL_CONF, "TimesGMT", FALSE);
   
-  /* No need to check ShowDotFiles since we force opt_a below.
-   */
   if(_fakemode != -1) {
-    fakemode = (umode_t)_fakemode;
+    fakemode = (mode_t)_fakemode;
     fakemodep = 1;
   } else
     fakemodep = 0;
 
+  if ((tmp = get_param_ptr(TOPLEVEL_CONF, "TimesGMT", FALSE)) != NULL)
+    list_times_gmt = *tmp;
+
   opt_C = opt_d = opt_F = opt_R;
   opt_a = opt_l = opt_STAT = 1;
 
@@ -1440,185 +1490,203 @@ MODRET cmd_list(cmd_rec *cmd)
  * matching the glob(s).
  */
 
-MODRET cmd_nlst(cmd_rec *cmd)
-{
-	char	*target,line[MAXPATHLEN + 1] = {'\0'};
-	int	count = 0,ret = 0, hidden = 0;
+MODRET cmd_nlst(cmd_rec *cmd) {
+  char *target,line[MAXPATHLEN + 1] = {'\0'};
+  int count = 0,ret = 0, hidden = 0;
+  unsigned char *tmp = NULL;
 
-	/* In case the client used NLST instead of LIST
-	 */
+  /* In case the client used NLST instead of LIST
+   */
+  if (cmd->argc > 1 && cmd->argv[1][0] == '-')
+    return genericlist(cmd);
 
-	if(cmd->argc > 1 && cmd->argv[1][0] == '-')
-		return genericlist(cmd);
-
-	showsymlinks = get_param_int(TOPLEVEL_CONF,"ShowSymlinks",FALSE);
-
-	if(showsymlinks == -1)
-		showsymlinks = 1;
-	if((opt_A = get_param_int(TOPLEVEL_CONF,"ShowDotFiles",FALSE)) < 0)
-	  opt_A = 0;
+  if ((tmp = get_param_ptr(TOPLEVEL_CONF, "ShowSymlinks", FALSE)) != NULL)
+    list_show_symlinks = *tmp;
 	
-	if(cmd->argc == 1)
-		target = ".";
-	else
-		target = cmd->arg;
+  if (cmd->argc == 1)
+    target = ".";
+  else
+    target = cmd->arg;
 	
-	/* If the target starts with '~' ... */
-	if(*target == '~') {
-		char pb[MAXPATHLEN + 1] = {'\0'};
-		struct passwd *pw;
-		int i;
-		const char *p;
+  /* If the target starts with '~' ... */
+  if(*target == '~') {
+    char pb[MAXPATHLEN + 1] = {'\0'};
+    struct passwd *pw = NULL;
+    int i = 0;
+    const char *p = target;
 
-		i = 0;
-		p = target;
-		p++;
+    p++;
 
-		while(*p && *p !='/' && i<MAXPATHLEN)
-			pb[i++] = *p++;
-		pb[i] = '\0';
+    while (*p && *p !='/' && i < MAXPATHLEN)
+      pb[i++] = *p++;
+    pb[i] = '\0';
 
-		if((pw = auth_getpwnam(cmd->tmp_pool,i ? pb : session.user))) {
-		  snprintf(pb, sizeof(pb), "%s%s",pw->pw_dir,p);
-		  sstrncpy(line, pb, sizeof(line));
-		  target = line;
-		}
-	}
+    if ((pw = auth_getpwnam(cmd->tmp_pool,i ? pb : session.user))) {
+      snprintf(pb, sizeof(pb), "%s%s", pw->pw_dir, p);
+      sstrncpy(line, pb, sizeof(line));
+      target = line;
+    }
+  }
 	
-	/* If the target is a glob, get the listing of files/dirs to send
-	 */
+  /* If the target is a glob, get the listing of files/dirs to send
+   */
+  if (use_globbing && strpbrk(target, "{[*?") != NULL) {
+    glob_t g;
+    char **path,*p;
 
-	if(strpbrk(target,"{[*?") != NULL) {
-		glob_t g;
-		char **path,*p;
-		
-		if(fs_glob(target,GLOB_PERIOD,NULL,&g) != 0) {
-			add_response_err(R_550,"No files found.");
-			return ERROR(cmd);
-		}
+    /* Make sure the glob_t is initialized */
+    memset(&g, '\0', sizeof(glob_t));
 
-		/* Iterate through each matching entry */
-		path = g.gl_pathv;
-		while(path && *path && ret >= 0) {
-			struct stat st;
-                        int hidden = 0;
+    if (fs_glob(target, GLOB_PERIOD,NULL, &g) != 0) {
+      add_response_err(R_550, "No files found");
+      return ERROR(cmd);
+    }
 
-			p = *path; path++;
+    /* Iterate through each matching entry */
+    path = g.gl_pathv;
+    while (path && *path && ret >= 0) {
+      struct stat st;
+      int hidden = 0;
 
-			if(*p == '.' && (!opt_A || is_dotdir(p)))
-				continue;
+      p = *path;
+      path++;
 
-			if(fs_stat(p,&st) == 0) {
-				/* If it's a directory, hand off to nlstdir */
-				if(S_ISDIR(st.st_mode))
-					ret = nlstdir(cmd,p);
-				else if(S_ISREG(st.st_mode) &&
-						ls_perms(cmd->tmp_pool,cmd,p,&hidden)) {
-                                        /* Don't display hidden files */
-                                        if(hidden)
-                                          continue;
+      if (*p == '.' && (!opt_A || is_dotdir(p)))
+        continue;
+
+      if (fs_stat(p, &st) == 0) {
+        /* If it's a directory, hand off to nlstdir */
+        if (S_ISDIR(st.st_mode))
+          ret = nlstdir(cmd, p);
+
+        else if (S_ISREG(st.st_mode) &&
+            ls_perms(cmd->tmp_pool, cmd, p, &hidden)) {
+          /* Don't display hidden files */
+          if (hidden)
+            continue;
                                         
-					ret = nlstfile(cmd,p);
-                                }
-				if(ret > 0)
-					count += ret;
-			}
-		}
-		fs_globfree(&g);
-	} else {
-		/* A single target, if it's a directory, list the contents,
-		 * if it's a file, just list the file.
-		 */
-		struct stat st;
+          ret = nlstfile(cmd,p);
+        }
+
+        if (ret > 0)
+          count += ret;
+      }
+    }
+
+    fs_globfree(&g);
+
+  } else {
+
+    /* A single target. If it's a directory, list the contents; if it's a
+     * file, just list the file.
+     */
+    struct stat st;
 		
-		if(!ls_perms(cmd->tmp_pool,cmd,target,&hidden)) {
-                    add_response_err(R_550,"%s: %s",cmd->arg,strerror(errno));
-		    return ERROR(cmd);
-		}
+    if (!ls_perms(cmd->tmp_pool, cmd, target, &hidden)) {
+      add_response_err(R_550, "%s: %s", cmd->arg, strerror(errno));
+      return ERROR(cmd);
+    }
 
+    /* Don't display hidden files */
+    if (hidden) {
+      config_rec *c = NULL;
+
+      if ((c = _find_ls_limit(target)) != NULL &&
+          get_param_int(c->subset, "IgnoreHidden", FALSE) == TRUE)
+        add_response_err(R_550, "%s: %s", cmd->arg, strerror(ENOENT));
+      else
+        add_response_err(R_550, "%s: %s", cmd->arg, strerror(EACCES));
+
+      return ERROR(cmd);
+    }
                 
-                /* Don't display hidden files */
-                if(hidden) {
-                  config_rec *c;
+    /* Make sure the target is a file or directory,
+     * and that we have access to it.
+     */
+    if (fs_stat(target, &st) < 0) {
+      add_response_err(R_550, "%s: %s", cmd->arg, strerror(errno));
+      return ERROR(cmd);
+    }
 
-                  if((c = _find_ls_limit(target)) != NULL &&
-                      get_param_int(c->subset, "IgnoreHidden", FALSE) == TRUE)
-                    add_response_err(R_550,"%s: %s", cmd->arg, strerror(ENOENT));
-                  else
-                    add_response_err(R_550,"%s: %s", cmd->arg, strerror(EACCES));
+    if (S_ISREG(st.st_mode))
+      ret = nlstfile(cmd, target);
 
-                  return ERROR(cmd);
-                }
-                
-		/* Make sure the target is a file or directory,
-		 * and that we have access to it.
-		 */
-		if(fs_stat(target,&st) < 0) {
-			add_response_err(R_550,"%s: %s",
-					cmd->arg,strerror(errno));
-			return ERROR(cmd);
-		}
+    else if (S_ISDIR(st.st_mode)) {
+      if (access_check(target, R_OK) != 0) {
+        add_response_err(R_550, "%s: %s", cmd->arg, strerror(errno));
+        return ERROR(cmd);
+      }
 
-		if(S_ISREG(st.st_mode))
-			ret = nlstfile(cmd,target);
-		else if(S_ISDIR(st.st_mode)) {
-			if(access_check(target,R_OK) != 0) {
-				add_response_err(R_550,"%s: %s",
-						cmd->arg,strerror(errno));
-				return ERROR(cmd);
-			}
-			ret = nlstdir(cmd,target);
-		} else {
-			add_response_err(R_550,"%s: Not a regular file",
-					cmd->arg,strerror(errno));
-			return ERROR(cmd);
-		}
+      ret = nlstdir(cmd,target);
+
+    } else {
+      add_response_err(R_550, "%s: Not a regular file", cmd->arg);
+      return ERROR(cmd);
+    }
 		
-		if(ret > 0)
-			count += ret;
-	}
-	
-	if(XFER_ABORTED) {
-		data_abort(0,0);
-		ret = -1;
-	} else {
-		if(ret == 0 && !count && (session.flags & SF_XFER) == 0) {
-			add_response_err(R_550,"No files found.");
-			ret = -1;
-		} else if(session.flags & SF_XFER)
-			ls_done(cmd);
+    if (ret > 0)
+      count += ret;
+  }
 
-		/* Note that the data connection is NOT cleared here,
-		 * as an error in NLST still leaves data ready for
-		 * another command
-		 */
-	}
+  if (XFER_ABORTED) {
+    data_abort(0, 0);
+    ret = -1;
 
-	return (ret < 0 ? ERROR(cmd) : HANDLED(cmd));
+  } else {
+    if (ret == 0 && !count && (session.flags & SF_XFER) == 0) {
+      add_response_err(R_550, "No files found");
+      ret = -1;
+
+    } else if (session.flags & SF_XFER)
+
+      /* Note that the data connection is NOT cleared here,
+       * as an error in NLST still leaves data ready for
+       * another command
+       */
+      ls_done(cmd);
+  }
+
+  return (ret < 0 ? ERROR(cmd) : HANDLED(cmd));
 }
 
-MODRET _sethide(cmd_rec *cmd, const char *param)
-{
-  int bool;
+MODRET ls_chk_glob(cmd_rec *cmd) {
+  long globbing = -1;
+
+  if ((globbing = get_param_int(TOPLEVEL_CONF, "UseGlobbing", FALSE)) != -1) {
+    if (!globbing) {
+      log_debug(DEBUG3, "UseGlobbing disabling globbing functionality");
+      use_globbing = FALSE; 
+    }
+  }
+
+  return DECLINED(cmd);
+}
+
+/* Configuration handlers
+ */
+
+MODRET _sethide(cmd_rec *cmd, const char *param) {
+  int bool = -1;
   char *as = "ftp";
   config_rec *c = NULL;
 
   CHECK_CONF(cmd,CONF_ROOT|CONF_VIRTUAL|CONF_ANON|CONF_GLOBAL|
     CONF_DIR|CONF_DYNDIR);
 
-  if(cmd->argc < 2 || cmd->argc > 3)
+  if (cmd->argc < 2 || cmd->argc > 3)
     CONF_ERROR(cmd,pstrcat(cmd->tmp_pool,"syntax: ",
                param," on|off [<id to display>]",NULL));
 
-  if((bool = get_boolean(cmd,1)) == -1)
+  if ((bool = get_boolean(cmd,1)) == -1)
      CONF_ERROR(cmd, "expected boolean argument");
 
-  if(bool == TRUE) {
+  if (bool == TRUE) {
     /* use the configured id to display rather than the default "ftp" */
     if(cmd->argc > 2)
       as = cmd->argv[2];
 
     c = add_config_param_str(param, 1, as);
+
   } else {
     /* still need to add a config_rec to turn off the display of fake ids */
     c = add_config_param_str(param, 0);
@@ -1629,75 +1697,98 @@ MODRET _sethide(cmd_rec *cmd, const char *param)
   return HANDLED(cmd);
 }
 
-MODRET set_dirfakeuser(cmd_rec *cmd)
-{
-  return _sethide(cmd,"DirFakeUser");
+MODRET set_dirfakeuser(cmd_rec *cmd) {
+  return _sethide(cmd, cmd->argv[0]);
 }
 
-MODRET set_dirfakegroup(cmd_rec *cmd)
-{
-  return _sethide(cmd,"DirFakeGroup");
+MODRET set_dirfakegroup(cmd_rec *cmd) {
+  return _sethide(cmd, cmd->argv[0]);
 }
 
-MODRET set_dirfakemode(cmd_rec *cmd)
-{
+MODRET set_dirfakemode(cmd_rec *cmd) {
   config_rec *c;
   unsigned long fake;
   char *endp;
 
-  CHECK_ARGS(cmd,1);
-  CHECK_CONF(cmd,CONF_ROOT|CONF_VIRTUAL|CONF_ANON|CONF_GLOBAL|CONF_DIR);
+  CHECK_ARGS(cmd, 1);
+  CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL|CONF_ANON|CONF_DIR);
 
   fake = (unsigned long)strtol(cmd->argv[1],&endp,8);
 
   if(endp && *endp)
     CONF_ERROR(cmd,"argument must be an octal number.");
 
-  c = add_config_param("DirFakeMode",1,(void*)fake);
+  c = add_config_param(cmd->argv[0], 1, (void *) fake);
   c->flags |= CF_MERGEDOWN;
 
   return HANDLED(cmd);
 }
 
-MODRET set_lsdefaultoptions(cmd_rec *cmd)
-{
-  CHECK_ARGS(cmd,1);
-  CHECK_CONF(cmd,CONF_ROOT|CONF_VIRTUAL|CONF_ANON|CONF_GLOBAL);
+MODRET set_lsdefaultoptions(cmd_rec *cmd) {
+  config_rec *c = NULL;
 
-  add_config_param_str("LsDefaultOptions",1,cmd->argv[1]);
+  CHECK_ARGS(cmd, 1);
+  CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL|CONF_ANON);
+
+  c = add_config_param_str(cmd->argv[0], 1, cmd->argv[1]);
+  c->flags |= CF_MERGEDOWN;
 
   return HANDLED(cmd);
 }
 
-MODRET set_showdotfiles(cmd_rec *cmd)
-{
-  int b;
+MODRET set_showsymlinks(cmd_rec *cmd) {
+  int bool = -1;
+  config_rec *c = NULL;
+
+  CHECK_ARGS(cmd, 1);
+  CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL|CONF_ANON);
+
+  if ((bool = get_boolean(cmd, 1)) == -1)
+    CONF_ERROR(cmd, "expected Boolean parameter");
+
+  c = add_config_param(cmd->argv[0], 1, NULL);
+  c->argv[0] = pcalloc(c->pool, sizeof(unsigned char));
+  *((unsigned char *) c->argv[0]) = bool;
+  c->flags |= CF_MERGEDOWN;
+
+  return HANDLED(cmd);
+}
+
+MODRET set_useglobbing(cmd_rec *cmd) {
+  int bool = -1;
+  config_rec *c = NULL;
   
-  CHECK_ARGS(cmd,1);
-  CHECK_CONF(cmd,CONF_ROOT|CONF_VIRTUAL|CONF_ANON|CONF_GLOBAL);
+  CHECK_ARGS(cmd, 1);
+  CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL|CONF_ANON);
 
-  if((b = get_boolean(cmd,1)) == -1)
-    CONF_ERROR(cmd,"expected boolean argument.");
+  if ((bool = get_boolean(cmd, 1)) == -1)
+    CONF_ERROR(cmd, "expected boolean parameter");
+
+  c = add_config_param(cmd->argv[0], 1, (void *) bool);
+  c->flags |= CF_MERGEDOWN;
 
-  add_config_param("ShowDotFiles",1,(void*)b);
   return HANDLED(cmd);
 }
 
-static conftable ls_config[] = {
+static conftable ls_conftab[] = {
   { "DirFakeUser",	set_dirfakeuser,			NULL },
   { "DirFakeGroup",	set_dirfakegroup,			NULL },
   { "DirFakeMode",	set_dirfakemode,			NULL },
   { "LsDefaultOptions",	set_lsdefaultoptions,			NULL },
-  { "ShowDotFiles",	set_showdotfiles,			NULL },
+  { "ShowSymlinks",	set_showsymlinks,			NULL },
+  { "UseGlobbing",	set_useglobbing,			NULL },
   { NULL,		NULL,					NULL }
 };
 
-static cmdtable ls_commands[] = {
+static cmdtable ls_cmdtab[] = {
   { CMD,  	C_NLST,	G_DIRS,	cmd_nlst,	TRUE, FALSE, CL_DIRS },
   { CMD,	C_LIST,	G_DIRS,	cmd_list,	TRUE, FALSE, CL_DIRS },
   { CMD, 	C_STAT,	G_DIRS,	cmd_stat,	TRUE, FALSE, CL_DIRS },
+  { POST_CMD,	C_PASS,	G_NONE,	ls_chk_glob,	FALSE, FALSE },
   { LOG_CMD,	C_LIST,	G_NONE,	fini_nlst,	FALSE, FALSE },
   { LOG_CMD,	C_NLST, G_NONE,	fini_nlst,	FALSE, FALSE },
+  { LOG_CMD_ERR,C_LIST, G_NONE, list_cleanup,   FALSE, FALSE },
+  { LOG_CMD_ERR,C_NLST, G_NONE, list_cleanup,   FALSE, FALSE },
   { 0, NULL }
 };
 
@@ -1705,8 +1796,9 @@ module ls_module = {
   NULL,NULL,			/* Always NULL */
   0x20,				/* API version */
   "ls",				/* Module name */
-  ls_config,
-  ls_commands,
+  ls_conftab,
+  ls_cmdtab,
   NULL,
-  NULL,NULL
+  NULL,
+  NULL
 };
