@@ -1,7 +1,8 @@
 /*
  * ProFTPD - FTP server daemon
- * Copyright (c) 1997, Public Flood Software
- *  
+ * Copyright (c) 1997, 1998 Public Flood Software
+ * Copyright (c) 2001, 2002 The ProFTPD Project team
+ *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
@@ -14,7 +15,12 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ *
+ * As a special exemption, Public Flood Software/MacGyver aka Habeeb J. Dihu
+ * and other respective copyright holders give permission to link this program
+ * with OpenSSL, and distribute the resulting executable, without including
+ * the source code for OpenSSL in the source distribution.
  */
 
 /* Simply utility to create the proftpd shutdown message file, allowing
@@ -33,6 +39,22 @@ static void show_usage(char *progname)
   exit(1);
 }
 
+static int isnumeric(char *str)
+{
+  while(str && isspace(*str))
+    str++;
+  
+  if(!str || !*str)
+    return 0;
+  
+  for(; str && *str; str++) {
+    if(!isdigit(*str))
+      return 0;
+  }
+  
+  return 1;
+}
+
 int main(int argc, char *argv[])
 {
   int deny = 10,disc = 5,c;
@@ -47,23 +69,28 @@ int main(int argc, char *argv[])
   while((c = getopt(argc,argv,"l:d:")) != -1) {
     switch(c) {
     case 'l':
-      if(!optarg) {
-        fprintf(stderr,"%s: -l requires an argument.\n",progname);
-        show_usage(progname);
-      }
-
-      deny = atoi(optarg);
-      break;
     case 'd':
       if(!optarg) {
-        fprintf(stderr,"%s: -d requires an argument.\n",progname);
+        fprintf(stderr,"%s: -%c requires an argument.\n", progname, c);
         show_usage(progname);
       }
 
-      deny = atoi(optarg);
+      if(!isnumeric(optarg)) {
+	fprintf(stderr, "%s: -%c requires a numeric argument.\n",
+		progname, c);
+	show_usage(progname);
+      }
+      
+      if(c == 'd')
+	disc = atoi(optarg);
+      else if(c == 'l')
+	deny = atoi(optarg);
+      
       break;
+      
     case '?':
       fprintf(stderr,"%s: unknown option '%c'.\n",progname,(char)optopt);
+      
     case 'h':
     default:
       show_usage(progname);
@@ -85,20 +112,44 @@ int main(int argc, char *argv[])
   tm = localtime(&now);
 
   /* shut must be either 'now', '+number' or 'HHMM' */
-  if(strcmp(shut,"now") != 0) {
+  if(strcasecmp(shut,"now") != 0) {
     if(*shut == '+') {
       shut++;
-      while(*shut && isspace(*shut)) shut++;
+      while(shut && *shut && isspace((UCHAR)*shut)) shut++;
 
+      if(!isnumeric(shut)) {
+	fprintf(stderr, "%s: Invalid time interval specified.\n", progname);
+	show_usage(progname);
+      }
+      
       now += (60 * atoi(shut));
       tm = localtime(&now);
     } else {
+      if((strlen(shut) != 4 && strlen(shut) != 2) || !isnumeric(shut)) {
+	fprintf(stderr, "%s: Invalid time interval specified.\n", progname);
+	show_usage(progname);
+      }
+
       if(strlen(shut) > 2) {
         mn = atoi((shut + strlen(shut) - 2));
+
+	if(mn > 59) {
+	  fprintf(stderr, "%s: Invalid time interval specified.\n",
+		  progname);
+	  show_usage(progname);
+	}
+
         *(shut + strlen(shut) - 2) = '\0';
       }
+
       hr = atoi(shut);
 
+      if(hr > 23) {
+	fprintf(stderr, "%s: Invalid time interval specified.\n",
+		progname);
+	show_usage(progname);
+      }
+      
       if(hr < tm->tm_hour || (hr == tm->tm_hour && mn <= tm->tm_min)) {
         now += 86400;		/* one day forward */
         tm = localtime(&now);
@@ -116,8 +167,8 @@ int main(int argc, char *argv[])
   }
 
   fprintf(outf,"%d %d %d %d %d %d",
-          tm->tm_year+1900,tm->tm_mon,tm->tm_mday,tm->tm_hour,
-          tm->tm_min,tm->tm_sec);
+          tm->tm_year+1900, tm->tm_mon + 1, tm->tm_mday, tm->tm_hour,
+          tm->tm_min, tm->tm_sec);
   fprintf(outf," %02d%02d %02d%02d\n",
           (deny / 60),(deny % 60),
           (disc / 60),(disc % 60));
